import warnings
from splink.settings import ComparisonColumn
from pyspark.sql.session import SparkSession

from copy import deepcopy

from .validate import get_default_value_from_schema

from .case_statements import (
    _check_jaro_registered,
    sql_gen_case_smnt_strict_equality_2,
    sql_gen_case_stmt_levenshtein_rel_3,
    sql_gen_case_stmt_levenshtein_rel_4,
    sql_gen_case_stmt_jaro_3,
    sql_gen_case_stmt_jaro_4,
    sql_gen_case_stmt_numeric_float_equality_2,
    sql_gen_case_stmt_numeric_perc_3,
    sql_gen_case_stmt_numeric_perc_4,
    _check_no_obvious_problem_with_case_statement,
    _add_as_gamma_to_case_statement,
)

from .parse_case_statement import (
    parse_case_statement,
    generate_sql_from_parsed_case_expr,
    get_columns_used_from_sql_without_l_r_suffix,
)


def _normalise_prob_list(prob_array: list):
    sum_list = sum(prob_array)
    return [i / sum_list for i in prob_array]


def _get_default_case_statements_functions(spark):
    default_case_stmts = {
        "numeric": {},
        "string": {},
    }

    default_case_stmts["numeric"][2] = sql_gen_case_stmt_numeric_float_equality_2
    default_case_stmts["numeric"][3] = sql_gen_case_stmt_numeric_perc_3
    default_case_stmts["numeric"][4] = sql_gen_case_stmt_numeric_perc_4

    if spark:
        jaro_exists = _check_jaro_registered(spark)
    else:
        jaro_exists = False

    if jaro_exists:
        default_case_stmts["string"][2] = sql_gen_case_smnt_strict_equality_2
        default_case_stmts["string"][3] = sql_gen_case_stmt_jaro_3
        default_case_stmts["string"][4] = sql_gen_case_stmt_jaro_4

    else:
        default_case_stmts["string"][2] = sql_gen_case_smnt_strict_equality_2
        default_case_stmts["string"][3] = sql_gen_case_stmt_levenshtein_rel_3
        default_case_stmts["string"][4] = sql_gen_case_stmt_levenshtein_rel_4

    return default_case_stmts


def _get_default_case_statement_fn(default_statements, data_type, levels):
    if data_type not in ["string", "numeric"]:
        raise ValueError(
            f"No default case statement available for data type {data_type}, "
            "please specify a custom case_expression"
        )
    if levels > 4:
        raise ValueError(
            f"No default case statement available when levels > 4, "
            "please specify a custom 'case_expression' within your settings dictionary"
        )
    return default_statements[data_type][levels]


def _get_default_probabilities(m_or_u, levels):

    if levels > 6:
        raise ValueError(
            f"No default m and u probabilities available when levels > 6, "
            "please specify custom values for 'm_probabilities' and 'u_probabilities' "
            "within your settings dictionary"
        )

    # Note all m and u probabilities are automatically normalised to sum to 1
    default_m_u_probabilities = {
        "m_probabilities": {
            2: _normalise_prob_list([1, 9]),
            3: _normalise_prob_list([1, 2, 7]),
            4: _normalise_prob_list([1, 1, 1, 7]),
            5: _normalise_prob_list([0.33, 0.67, 1, 2, 6]),
            6: _normalise_prob_list([0.33, 0.67, 1, 2, 3, 6]),
        },
        "u_probabilities": {
            2: _normalise_prob_list([9, 1]),
            3: _normalise_prob_list([7, 2, 1]),
            4: _normalise_prob_list([7, 1, 1, 1]),
            5: _normalise_prob_list([6, 2, 1, 0.33, 0.67]),
            6: _normalise_prob_list([6, 3, 2, 1, 0.33, 0.67]),
        },
    }

    probabilities = default_m_u_probabilities[m_or_u][levels]
    return probabilities


def _complete_case_expression(col_settings, spark):

    cc = ComparisonColumn(col_settings)
    if cc.has_case_expression_or_comparison_levels:
        return col_settings

    default_case_statements = _get_default_case_statements_functions(spark)
    levels = cc.num_levels

    if "custom_name" in col_settings:
        col_name_for_case_fn = col_settings["custom_name"]
    else:
        col_name_for_case_fn = col_settings["col_name"]

    if "case_expression" not in col_settings:
        data_type = col_settings["data_type"]
        case_fn = _get_default_case_statement_fn(
            default_case_statements, data_type, levels
        )
        col_settings["case_expression"] = case_fn(
            col_name_for_case_fn, col_name_for_case_fn
        )
    else:
        _check_no_obvious_problem_with_case_statement(col_settings["case_expression"])
        old_case_stmt = col_settings["case_expression"]
        new_case_stmt = _add_as_gamma_to_case_statement(
            old_case_stmt, col_name_for_case_fn
        )
        col_settings["case_expression"] = new_case_stmt


def _complete_probabilities(col_settings: dict, mu_probabilities: str):
    """

    Args:
        col_settings (dict): Column settings dictionary
        mu_probabilities (str): Either 'm_probabilities' or 'u_probabilities'

    """

    if mu_probabilities not in col_settings:
        cc = ComparisonColumn(col_settings)
        levels = cc.num_levels
        probs = _get_default_probabilities(mu_probabilities, levels)
        col_settings[mu_probabilities] = probs


def _complete_tf_adjustment_weights(col_settings: dict):

    if "tf_adjustment_weights" in col_settings:
        if not all(0.0 <= w <= 1.0 for w in col_settings["tf_adjustment_weights"]):
            raise ValueError(
                f"All values of 'tf_adjustment_weights' must be between 0 and 1"
            )
    else:
        cc = ComparisonColumn(col_settings)

        weights = [0.0] * cc.num_levels
        weights[-1] = 1.0
        col_settings["tf_adjustment_weights"] = weights


def _complete_comparison_levels(col_settings):
    if "comparison_levels" not in col_settings:
        case_expression = col_settings["case_expression"]
        col_settings["comparison_levels"] = parse_case_statement(case_expression)

    if "case_expression" not in col_settings:
        cl = col_settings["comparison_levels"]
        col_settings["case_expression"] = generate_sql_from_parsed_case_expr(cl)

    from splink.settings import ComparisonColumn

    cc = ComparisonColumn(col_settings)
    keys = cc.comparison_levels_dict.keys()
    if "-1" not in keys:

        warnings.warn(
            "\nNo -1 level found in case statement."
            " You usually want to use -1 as the level for the null value."
            " e.g. WHEN col_l is null or col_r is null then -1"
            f" Case statement is:\n {col_settings['case_expression']}."
        )

    if "num_levels" in col_settings:
        if col_settings["num_levels"] != cc.num_levels:
            warnings.warn(
                f"\nnum_levels specified in settings is {col_settings['num_levels']}, "
                f"but the number of levels in the case statement excluding the null level is {cc.num_levels}"
                f"\nCase statement is:\n {col_settings['case_expression']}."
            )


def _complete_col_name(col_settings):

    if "custom_name" in col_settings:
        return

    if "col_name" in col_settings:
        return

    sql = generate_sql_from_parsed_case_expr(col_settings["comparison_levels"])
    sql_cols = get_columns_used_from_sql_without_l_r_suffix(sql)
    if len(sql_cols) == 1:
        col_settings["col_name"] = sql_cols[0]
    else:
        col_settings["custom_name"] = "_".join(sql_cols)
    return col_settings


def _complete_custom_columns(col_settings):

    if "col_name" in col_settings:
        return col_settings

    if "custom_name" in col_settings:
        sql = generate_sql_from_parsed_case_expr(col_settings["comparison_levels"])
        sql_cols = get_columns_used_from_sql_without_l_r_suffix(sql)
        if "columns_used" in col_settings:
            if set(sql_cols) != set(col_settings["columns_used"]):
                warnings.warn(
                    f"The columns used in the case statement are {sql_cols} but the columns "
                    f"specified in the settings dictionary are {col_settings['columns_used']}"
                )
        else:
            col_settings["custom_columns_used"] = sql_cols
    return col_settings


def complete_settings_dict(settings_dict: dict, spark: SparkSession):
    """Auto-populate any missing settings from the settings dictionary using the 'sensible defaults' that
    are specified in the json schema (./splink/files/settings_jsonschema.json)

    Args:
        settings_dict (dict): The settings dictionary
        spark: The SparkSession

    Returns:
        dict: A `splink` settings dictionary with all keys populated.
    """
    settings_dict = deepcopy(settings_dict)

    # Complete non-column settings from their default values if not exist
    non_col_keys = [
        "link_type",
        "em_convergence",
        "source_dataset_column_name",
        "unique_id_column_name",
        "additional_columns_to_retain",
        "retain_matching_columns",
        "retain_intermediate_calculation_columns",
        "max_iterations",
        "proportion_of_matches",
    ]
    for key in non_col_keys:
        if key not in settings_dict:
            settings_dict[key] = get_default_value_from_schema(
                key, is_column_setting=False
            )

    if "blocking_rules" in settings_dict:
        if len(settings_dict["blocking_rules"]) == 0:
            warnings.warn(
                "You have not specified any blocking rules, meaning all comparisons between the "
                "input dataset(s) will be generated and blocking will not be used."
                "For large input datasets, this will generally be computationally intractable "
                "because it will generate comparisons equal to the number of rows squared."
            )

    c_cols = settings_dict["comparison_columns"]
    for gamma_index, col_settings in enumerate(c_cols):

        # Gamma index refers to the position in the comparison vector
        # i.e. it's a counter for comparison columns
        col_settings["gamma_index"] = gamma_index

        # Populate non-existing keys from defaults
        keys_for_defaults = [
            "data_type",
            "term_frequency_adjustments",
            "fix_u_probabilities",
            "fix_m_probabilities",
        ]

        for key in keys_for_defaults:
            if key not in col_settings:
                default = get_default_value_from_schema(key, is_column_setting=True)
                col_settings[key] = default

        # Populate default value for num levels only if case_expression or comparison_levels is not specified
        skip_if_present = set(["case_expression", "comparison_levels", "num_levels"])
        keys = set(col_settings.keys())
        intersect = keys.intersection(skip_if_present)
        if len(intersect) == 0:
            default = get_default_value_from_schema(
                "num_levels", is_column_setting=True
            )
            col_settings["num_levels"] = default

        # Doesn't need assignment because we're modify the col_settings dictionary

        _complete_case_expression(col_settings, spark)

        _complete_comparison_levels(col_settings)
        _complete_col_name(col_settings)
        _complete_custom_columns(col_settings)

        _complete_probabilities(col_settings, "m_probabilities")
        _complete_probabilities(col_settings, "u_probabilities")

        _complete_tf_adjustment_weights(col_settings)

    return settings_dict


def normalise_probabilities(settings_dict: dict):
    """Normalise all probabilities in a settings dictionary to sum
    to one, of possible

    Args:
        settings_dict (dict): Splink settings dictionary
    """

    c_cols = settings_dict["comparison_columns"]
    for col_settings in c_cols:
        for p in ["m_probabilities", "u_probabilities"]:
            if p in col_settings:
                if None not in col_settings[p]:
                    if sum(col_settings[p]) != 0:
                        col_settings[p] = _normalise_prob_list(col_settings[p])
    return settings_dict



from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
import pyspark.sql.functions as f

from splink import Splink
from .blocking import (
    sql_gen_comparison_columns, 
    _get_columns_to_retain_blocking
)
from .model import Model
from .validate import validate_input_datasets
from .gammas import _sql_gen_add_gammas
from .iterate import iterate
from .charts import altair_if_installed_else_json, load_chart_definition

def _self_link(linker: Splink):
    """Return a DataFrame of edges between every record in linker.df and itself
    (i.e. the highest possible splink score for each record)

    Args:
        linker (Splink): A Splink data linker
    """    

    # Take df and settings from Splink object
    df = linker.df
    settings = linker.settings_dict
    spark = linker.spark

    unique_id_col = settings["unique_id_column_name"]

    # Change settings to link each record in df to itself
    settings["link_type"] = "link_only"
    settings["max_iterations"] = 0
    settings["retain_matching_columns"] = True
    settings["blocking_rules"] = [f'l.{unique_id_col} = r.{unique_id_col}']

    model = Model(settings, spark)

    df.createOrReplaceTempView("df")

    columns_to_retain = _get_columns_to_retain_blocking(settings, df)
    columns_to_retain.remove(settings["source_dataset_column_name"])

    # Fudge df_comparisons (because block_using_rules doesn't quite work)
    sql = f"""select
        {sql_gen_comparison_columns(columns_to_retain)},
        '0' as match_key
        from df as l
        inner join df as r
        on {settings['blocking_rules'][0]}
    """

    df_comparison = spark.sql(sql)\
        .withColumn("source_dataset_l", f.lit("l"))\
        .withColumn("source_dataset_r", f.lit("r"))
    df_comparison.createOrReplaceTempView("df_comparison")

    sql = _sql_gen_add_gammas(settings, df_comparison)
    df_gammas = spark.sql(sql)

    # df_e with same no. of rows as df, with a self-link score for each record
    df_e = iterate(df_gammas, model, spark)

    return df_e

def unlinkables_chart(linker: Splink, x_col="match_weight", source_dataset=None):
    """Generate a chart displaying the proportion of records that are "unlinkable"
    for a given splink score threshold and model parameters. These are records that,
    even when compared with themselves, do not contain enough information to confirm
    a match.

    Args:
        linker (Splink): A Splink data linker
        x_col (str, optional): The column name to use as the x-axis in the chart.
            This can be either the "match_weight" or "match_probability" columns. 
            Defaults to "match_weight".
        source_dataset (str, optional): Name of the source dataset (used in chart 
            title). Defaults to None.
    """    

    df_self = _self_link(linker)
    
    if x_col not in ["match_weight", "match_probability"]:
        raise ValueError(
            f"{x_col} must be 'match_weight' (default) or 'match_probability'."
        )

    chart_path = "unlinkables_chart_def.json"
    unlinkables_chart_def = load_chart_definition(chart_path)


    data = df_self.groupBy(f.round("match_probability",5).alias("match_probability"))\
        .agg(
            f.count("match_weight").alias("count"), 
            f.max(f.round("match_weight",2)).alias("match_weight")
        ).toPandas()
    data = data.sort_values("match_probability").reset_index(drop=True)
    data["prop"]= data["count"]/ data["count"].sum()
    data["cum_prop"] = data["prop"].cumsum()
    data = data[:-1]

    unlinkables_chart_def["data"]["values"] = data.to_dict('records')

    if x_col == "match_probability":
        unlinkables_chart_def["layer"][0]["encoding"]["x"]["field"] = 'match_probability'
        unlinkables_chart_def["layer"][0]["encoding"]["x"]["axis"]["title"] = 'Threshold match probability'
        unlinkables_chart_def["layer"][0]["encoding"]["x"]["axis"]["format"] = '.2'

        unlinkables_chart_def["layer"][1]["encoding"]["x"]["field"] = 'match_probability'
        unlinkables_chart_def["layer"][1]["selection"]["selector112"]["fields"] = ['match_probability', 'cum_prop']

        unlinkables_chart_def["layer"][2]["encoding"]["x"]["field"] = 'match_probability'
        unlinkables_chart_def["layer"][2]["encoding"]["x"]["axis"]["title"] = 'Threshold match probability'

        unlinkables_chart_def["layer"][3]["encoding"]["x"]["field"] = 'match_probability'

    if source_dataset:
        unlinkables_chart_def["title"]["text"] += f" - {source_dataset}"

    return altair_if_installed_else_json(unlinkables_chart_def)


from functools import reduce
import pyspark.sql.functions as f
from pyspark.sql import Window
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
import pyspark
from typing import Union

altair_installed = True
try:
    import altair as alt
except ImportError:
    altair_installed = False


def _sql_gen_unique_id_keygen(
    table: str,
    uid_col1: str,
    uid_col2: str,
    source_dataset1: str = None,
    source_dataset2: str = None,
):
    """Create a composite unique id for a pairwise comparisons
    This is a concatenation of the unique id of each record
    of the pairwise comparisons.

    The composite unique id is agnostic to the ordering
    i.e. it treats:
    unique_id_l = x, unique_id_r = y
    and
    unique_id_l = y, unique_id_r = x
    to be equivalent.

    This is necessarily because we cannot predict
    which way round they will appear

    Args:
        table (str): name of table
        uid_col1 (str): name of unique id col 1
        uid_col2 (str): name of unique id col 2
        source_dataset1 (str, optional): Name of source dataset column if exists. Defaults to None.
        source_dataset2 (str, optional): [description]. Defaults to None.

    Returns:
        str: sql case expression that outputs the composite unique_id
    """

    if source_dataset1:
        concat_1 = f"concat({table}.{source_dataset1}, '_',{table}.{uid_col1})"
        concat_2 = f"concat({table}.{source_dataset2}, '_',{table}.{uid_col2})"
    else:
        concat_1 = f"{table}.{uid_col1}"
        concat_2 = f"{table}.{uid_col2}"

    return f"""
    case
    when {concat_1} > {concat_2}
        then concat({concat_2}, '-', {concat_1})
    else concat({concat_1}, '-', {concat_2})
    end
    """


def _get_score_colname(df_e, score_colname=None):
    if score_colname:
        return score_colname
    elif "match_probability" in df_e.columns:
        return "match_probability"
    else:
        raise ValueError("There doesn't appear to be a score column in df_e")


def dedupe_splink_scores(
    df_e_with_dupes: DataFrame,
    unique_id_colname: str,
    score_colname: str = None,
    selection_fn: str = "abs_val",
):
    """Sometimes, multiple Splink jobs with different blocking rules are combined
    into a single dataset of edges.  Sometimes,the same pair of nodes will be
    scored multiple times, once by each job.  We need to deduplicate this dataset
    so each pair of nodes appears only once

    Args:
        df_e_with_dupes (DataFrame): Dataframe with dupes
        unique_id_colname (str): Unique id column name e.g. unique_id
        score_colname (str, optional): Which column contains scores? If none, inferred from
            df_e_with_dupes.columns. Defaults to None.
        selection_fn (str, optional): Where we have several different scores for a given
            pair of records, how do we decide the final score?
            Options are 'abs_val' and 'mean'.
            abs_val:  Take the value furthest from 0.5 i.e. the value that expresses most certainty
            mean: Take the mean of all values
            Defaults to 'abs_val'.
    """

    # Looking in blocking.py, the position of unique ids
    # (whether they appear in _l or _r) is guaranteed
    # in blocking outputs so we don't need to worry about
    # inversions

    # This is not the case for labelled data - hence the need
    # _sql_gen_unique_id_keygen to join labels to df_e

    possible_vals = ["abs_val", "mean"]
    if selection_fn not in possible_vals:
        raise ValueError(
            f"selection function should be in {possible_vals}, you passed {selection_fn}"
        )

    score_colname = _get_score_colname(df_e_with_dupes, score_colname)

    if selection_fn == "abs_val":
        df_e_with_dupes = df_e_with_dupes.withColumn(
            "absval", f.expr(f"0.5 - abs({score_colname})")
        )

        win_spec = Window.partitionBy(
            [f"{unique_id_colname}_l", f"{unique_id_colname}_r"]
        ).orderBy(f.col("absval").desc())
        df_e_with_dupes = df_e_with_dupes.withColumn(
            "ranking", f.row_number().over(win_spec)
        )
        df_e = df_e_with_dupes.filter(f.col("ranking") == 1)
        df_e = df_e.drop("absval")
        df_e = df_e.drop("ranking")

    if selection_fn == "mean":

        win_spec = Window.partitionBy(
            [f"{unique_id_colname}_l", f"{unique_id_colname}_r"]
        ).orderBy(f.col(score_colname).desc())

        df_e_with_dupes = df_e_with_dupes.withColumn(
            "ranking", f.row_number().over(win_spec)
        )

        df_e_with_dupes = df_e_with_dupes.withColumn(
            score_colname,
            f.avg(score_colname).over(
                win_spec.rowsBetween(
                    Window.unboundedPreceding, Window.unboundedFollowing
                )
            ),
        )
        df_e = df_e_with_dupes.filter(f.col("ranking") == 1)

        df_e = df_e.drop("ranking")

    return df_e


def labels_with_splink_scores(
    df_labels,
    df_e,
    unique_id_colname,
    spark,
    score_colname=None,
    source_dataset_colname=None,
    retain_all_cols=False,
):
    """Create a dataframe with clerical labels set against splink scores

    Assumes uniqueness of pairs of identifiers in both datasets - e.g.
    if you have duplicate clerical labels or splink scores, you should
    deduplicate them first

    Args:
        df_labels:  a dataframe like:
             | unique_id_l | unique_id_r | clerical_match_score |
             |:------------|:------------|---------------------:|
             | id1         | id2         |                  0.9 |
             | id1         | id3         |                  0.1 |
        df_e: a dataframe like
             | unique_id_l| unique_id_r| match_probability      |
             |:-----------|:-----------|-----------------------:|
             | id1        | id2        |                   0.85 |
             | id1        | id3        |                   0.2  |
             | id2        | id3        |                   0.1  |
        unique_id_colname (str): Unique id column name e.g. unique_id
        spark : SparkSession
        score_colname (float, optional): Allows user to explicitly state the column name
            in the Splink dataset containing the Splink score.  If none will be inferred
        source_dataset_colname (str, optional): if your datasets contain a source_dataset column, what name is it?
            and set join_on_source_dataset=True, which will include the source dataset in the join key Defaults to False.
        retain_all_cols (bool, optional): Retain all columns in input datasets. Defaults to False.

    Returns:
        DataFrame: Like:
             |   unique_id_l |   unique_id_r |   clerical_match_score |   match_probability      | found_by_blocking   |
             |--------------:|--------------:|-----------------------:|-------------------------:|:--------------------|
             |             0 |             1 |                      1 |                 0.999566 | True                |
             |             0 |             2 |                      1 |                 0.999566 | True                |
             |             0 |             3 |                      1 |                 0.999989 | True                |

    """
    score_colname = _get_score_colname(df_e, score_colname)

    uid_col_l = f"{unique_id_colname}_l"
    uid_col_r = f"{unique_id_colname}_r"

    df_labels.createOrReplaceTempView("df_labels")
    df_e.createOrReplaceTempView("df_e")

    if source_dataset_colname:
        source_ds_col_l = f"{source_dataset_colname}_l"
        source_ds_col_r = f"{source_dataset_colname}_r"
        labels_key = _sql_gen_unique_id_keygen(
            "df_labels", uid_col_l, uid_col_r, source_ds_col_l, source_ds_col_r
        )
        df_e_key = _sql_gen_unique_id_keygen(
            "df_e", uid_col_l, uid_col_r, source_ds_col_l, source_ds_col_r
        )
        select_cols = f"""
          coalesce(df_e.{source_ds_col_l},df_labels.{source_ds_col_l}) as {source_ds_col_l},
          coalesce(df_e.{uid_col_l},df_labels.{uid_col_l}) as {uid_col_l},
          coalesce(df_e.{source_ds_col_r},df_labels.{source_ds_col_r}) as {source_ds_col_r},
          coalesce(df_e.{uid_col_r},df_labels.{uid_col_r}) as {uid_col_r}
        """
    else:
        labels_key = _sql_gen_unique_id_keygen("df_labels", uid_col_l, uid_col_r)
        df_e_key = _sql_gen_unique_id_keygen("df_e", uid_col_l, uid_col_r)
        select_cols = f"""
          coalesce(df_e.{uid_col_l},df_labels.{uid_col_l}) as {uid_col_l},
          coalesce(df_e.{uid_col_r},df_labels.{uid_col_r}) as {uid_col_r}
        """

    if retain_all_cols:
        cols1 = [f"df_e.{c} as df_e__{c}" for c in df_e.columns if c != score_colname]
        cols2 = [
            f"df_labels.{c} as df_labels__{c}"
            for c in df_labels.columns
            if c != "clerical_match_score"
        ]
        cols1.extend(cols2)

        select_smt = ", ".join(cols1)

        sql = f"""
        select

        {select_smt},
        clerical_match_score,

        case
        when {score_colname} is null then 0
        else {score_colname}
        end as {score_colname},

        case
        when {score_colname} is null then false
        else true
        end as found_by_blocking


        from df_labels
        left join df_e
        on {labels_key}
        = {df_e_key}

        """

    else:
        sql = f"""
        select

        {select_cols},
        clerical_match_score,

        case
        when {score_colname} is null then 0
        else {score_colname}
        end as {score_colname},

        case
        when {score_colname} is null then false
        else true
        end as found_by_blocking


        from df_labels
        left join df_e
        on {labels_key}
        = {df_e_key}

        """
    return spark.sql(sql)


def _summarise_truth_cats(df_truth_cats, spark):

    df_truth_cats.createOrReplaceTempView("df_truth_cats")

    sql = """

    select
    avg(truth_threshold) as truth_threshold,
    count(*) as row_count,
    sum(cast(P as int)) as P,
    sum(cast(N as int)) as N,
    sum(cast(TP as int)) as TP,
    sum(cast(TN as int)) as TN,
    sum(cast(FP as int)) as FP,
    sum(cast(FN as int)) as FN

    from df_truth_cats
    """

    df_truth_cats = spark.sql(sql)

    df_truth_cats.createOrReplaceTempView("df_truth_cats")

    sql = f"""

    select
    *,
    P/row_count as P_rate,
    N/row_count as N_rate,
    TP/P as TP_rate,
    TN/N as TN_rate,
    FP/N as FP_rate,
    FN/P as FN_rate,
    TP/(TP+FP) as precision,
    TP/(TP+FN) as recall

    from df_truth_cats
    """

    return spark.sql(sql)


def df_e_with_truth_categories(
    df_labels_with_splink_scores,
    threshold_pred,
    spark: SparkSession,
    threshold_actual: float = 0.5,
    score_colname: str = None,
):
    """Join Splink's predictions to clerically labelled data and categorise
    rows by truth category (false positive, true positive etc.)

    Note that df_labels

    Args:
        df_labels_with_splink_scores (DataFrame): A dataframe of labels and associated splink scores
            usually the output of the truth.labels_with_splink_scores function
        threshold_pred (float): Threshold to use in categorising Splink predictions into
            match or no match
        spark (SparkSession): SparkSession object
        threshold_actual (float, optional): Threshold to use in categorising clerical match
            scores into match or no match. Defaults to 0.5.
        score_colname (float, optional): Allows user to explicitly state the column name
            in the Splink dataset containing the Splink score.  If none will be inferred

    Returns:
        DataFrame: Dataframe of labels associated with truth category
    """

    df_labels_with_splink_scores.createOrReplaceTempView("df_labels_with_splink_scores")

    score_colname = _get_score_colname(df_labels_with_splink_scores)

    pred = f"({score_colname} >= {threshold_pred})"

    actual = f"(clerical_match_score >= {threshold_actual})"

    sql = f"""
    select
    *,
    cast ({threshold_pred} as float) as truth_threshold,
    {actual} = 1.0 as P,
    {actual} = 0.0 as N,
    {pred} = 1.0 and {actual} = 1.0 as TP,
    {pred} = 0.0 and {actual} = 0.0 as TN,
    {pred} = 1.0 and {actual} = 0.0 as FP,
    {pred} = 0.0 and {actual} = 1.0 as FN

    from
    df_labels_with_splink_scores

    """

    return spark.sql(sql)


def _truth_space_table_old(
    df_labels_with_splink_scores: DataFrame,
    spark: SparkSession,
    threshold_actual: float = 0.5,
    score_colname: str = None,
):
    """Create a table of the ROC space i.e. truth table statistics
    for each discrimination threshold

    Args:
        df_labels_with_splink_scores (DataFrame): A dataframe of labels and associated splink scores
            usually the output of the truth.labels_with_splink_scores function
        threshold_actual (float, optional): Threshold to use in categorising clerical match
            scores into match or no match. Defaults to 0.5.
        score_colname (float, optional): Allows user to explicitly state the column name
            in the Splink dataset containing the Splink score.  If none will be inferred

    Returns:
        DataFrame: Table of 'truth space' i.e. truth categories for each threshold level
    """

    # This is used repeatedly to generate the roc curve
    df_labels_with_splink_scores.persist()

    # We want percentiles of score to compute
    score_colname = _get_score_colname(df_labels_with_splink_scores, score_colname)

    percentiles = [x / 100 for x in range(0, 101)]

    values_distinct = df_labels_with_splink_scores.select(score_colname).distinct()
    thresholds = values_distinct.stat.approxQuantile(score_colname, percentiles, 0.0)
    thresholds.append(1.01)
    thresholds = sorted(set(thresholds))

    roc_dfs = []
    for thres in thresholds:
        df_e_t = df_e_with_truth_categories(
            df_labels_with_splink_scores, thres, spark, threshold_actual, score_colname
        )
        df_roc_row = _summarise_truth_cats(df_e_t, spark)
        roc_dfs.append(df_roc_row)

    all_roc_df = reduce(DataFrame.unionAll, roc_dfs)
    return all_roc_df


def truth_space_table(
    df_labels_with_splink_scores: DataFrame,
    spark: SparkSession,
    threshold_actual: float = 0.5,
    score_colname: str = None,
):
    """Create a table of the ROC space i.e. truth table statistics
    for each discrimination threshold

    Args:
        df_labels_with_splink_scores (DataFrame): A dataframe of labels and associated splink scores
            usually the output of the truth.labels_with_splink_scores function
        threshold_actual (float, optional): Threshold to use in categorising clerical match
            scores into match or no match. Defaults to 0.5.
        score_colname (float, optional): Allows user to explicitly state the column name
            in the Splink dataset containing the Splink score.  If none will be inferred

    Returns:
        DataFrame: Table of 'truth space' i.e. truth categories for each threshold level
    """

    # At a truth threshold of 1.0, we say a splink score of 1.0 is a positive in ROC space. i.e it's inclusive, so if there are splink scores of exactly 1.0 it's not possible to have zero positives in the truth table.
    # This means that at a truth threshold of 0.0 we say a splink score of 0.0 positive.  so it's possible to have zero negatives in the truth table.

    # This code provides an efficient way to compute the truth space
    # It's more complex than the previous code, but executes much faster because only a single SQL query/PySpark Action is needed
    # The previous implementation, which is easier to understand, is [here](https://github.com/moj-analytical-services/splink/blob/b4f601e6d180c6abfd64ab40775dca3e3513c0b5/splink/truth.py#L396)

    # We start with df_labels_with_splink_scores
    # This is a table of each pair of clerically labelled records accompanied by the Splink match score.
    # It is sorted in order to clerical_match_score, low to high

    # This means for any row, if we picked a threshold equal to clerical_match_score, all rows _above_ (in the table) are categoried by splink as non-matches.

    # For instance, if a clerical_match_score is 0.25, then any records above this in the table have a score of <0.25, and are therefore negative.  We categorise a score of exactly 0.25 as positive.

    # In addition, we can categorise any indiviual row as containing a false positive or false negative _at_ the clerical match score for the row.

    # This allows us to say things like:  Of the records above this row, we have _classified_ them all as negative, but we have _seen_ to true (clerically labelled) positives.  Thefore these must be false negatives.

    # In particular, the calculations are as follows:
    # False positives:  The cumulative total of positive labels in records BELOW this row, INCLUSIVE (because this one is being counted as positive)
    # True positives:  The total number of positives minus false positives

    # False negatives:  The total number of negatives, minus negatives seen above this row
    # True negatives:  The cumulative total of negative labels in records aboev this row

    # We want percentiles of score to compute
    score_colname = _get_score_colname(df_labels_with_splink_scores, score_colname)

    df_labels_with_splink_scores.createOrReplaceTempView("df_labels_with_splink_scores")
    sql = f"""
    select
    *,
    {score_colname} as truth_threshold,
    case when clerical_match_score >= {threshold_actual} then 1
    else 0
    end
    as c_P,
    case when clerical_match_score >= {threshold_actual} then 0
    else 1
    end
    as c_N
    from df_labels_with_splink_scores
    order by {score_colname}
    """
    df_with_labels = spark.sql(sql)
    df_with_labels.createOrReplaceTempView("df_with_labels")

    sql = """
    select truth_threshold, count(*) as num_records_in_row, sum(c_P) as c_P, sum(c_N) as c_N
    from
    df_with_labels
    group by truth_threshold
    order by truth_threshold
    """
    df_with_labels_grouped = spark.sql(sql)
    df_with_labels_grouped.createOrReplaceTempView("df_with_labels_grouped")

    sql = """
    select
    truth_threshold,

    (sum(c_P) over (order by truth_threshold desc))  as cum_clerical_P,
    (sum(c_N) over (order by truth_threshold)) - c_N as cum_clerical_N,

    (select sum(c_P) from df_with_labels_grouped) as total_clerical_P,
    (select sum(c_N) from df_with_labels_grouped) as total_clerical_N,
    (select sum(num_records_in_row) from df_with_labels_grouped) as row_count,

    -num_records_in_row + sum(num_records_in_row) over (order by truth_threshold) as N_labels,
    sum(num_records_in_row) over (order by truth_threshold desc) as P_labels
    from df_with_labels_grouped
    order by  truth_threshold
    """
    df_with_cumulative_labels = spark.sql(sql)
    df_with_cumulative_labels.createOrReplaceTempView("df_with_cumulative_labels")

    sql = """
    select
    truth_threshold,
    row_count,
    total_clerical_P as P,
    total_clerical_N as N,

    P_labels - cum_clerical_P as FP,
    cum_clerical_P as TP,

    N_labels - cum_clerical_N as FN,
    cum_clerical_N as TN

    from df_with_cumulative_labels
    """
    df_with_truth_cats = spark.sql(sql)
    df_with_truth_cats.createOrReplaceTempView("df_with_truth_cats")
    df_with_truth_cats.toPandas()

    sql = """
    select
    truth_threshold,
    row_count,
    P,
    N,
    TP,
    TN,
    FP,
    FN,
    P/row_count as P_rate,
        N/row_count as N_rate,
        TP/P as TP_rate,
        TN/N as TN_rate,
        FP/N as FP_rate,
        FN/P as FN_rate,
        TP/(TP+FP) as precision,
        TP/(TP+FN) as recall
    from df_with_truth_cats
    """
    df_truth_space = spark.sql(sql)

    return df_truth_space


def roc_chart(
    df_labels_with_splink_scores: Union[DataFrame, dict],
    spark: SparkSession,
    threshold_actual: float = 0.5,
    x_domain: list = None,
    width: int = 400,
    height: int = 400,
):
    """Create a ROC chart from labelled data

    Args:
        df_labels_with_splink_scores (Union[DataFrame, dict]): A dataframe of labels and associated splink scores
            usually the output of the truth.labels_with_splink_scores function.  Or, a dict containing
            one such dataframe per key.  {'model 1': df1, 'model 2': df2}.  If a dict is provided, the
            ROC charts for each model will be plotted on the same figure.
        spark (SparkSession): SparkSession object
        threshold_actual (float, optional): Threshold to use in categorising clerical match
            scores into match or no match. Defaults to 0.5.
        x_domain (list, optional): Domain for x axis. Defaults to None.
        width (int, optional):  Defaults to 400.
        height (int, optional):  Defaults to 400.

    """

    roc_chart_def = {
        "config": {"view": {"continuousWidth": 400, "continuousHeight": 300}},
        "data": {"name": "data-fadd0e93e9546856cbc745a99e65285d", "values": None},
        "mark": {"type": "line", "clip": True, "point": True},
        "encoding": {
            "tooltip": [
                {"type": "quantitative", "field": "truth_threshold"},
                {"type": "quantitative", "field": "FP_rate"},
                {"type": "quantitative", "field": "TP_rate"},
                {"type": "quantitative", "field": "TP"},
                {"type": "quantitative", "field": "TN"},
                {"type": "quantitative", "field": "FP"},
                {"type": "quantitative", "field": "FN"},
                {"type": "quantitative", "field": "precision"},
                {"type": "quantitative", "field": "recall"},
            ],
            "x": {
                "type": "quantitative",
                "field": "FP_rate",
                "sort": ["truth_threshold"],
                "title": "False Positive Rate amongst clerically reviewed records",
            },
            "y": {
                "type": "quantitative",
                "field": "TP_rate",
                "sort": ["truth_threshold"],
                "title": "True Positive Rate amongst clerically reviewed records",
            },
            "color": {
                "type": "nominal",
                "field": "roc_label",
            },
        },
        "selection": {
            "selector076": {
                "type": "interval",
                "bind": "scales",
                "encodings": ["x"],
            }
        },
        "height": height,
        "title": "Receiver operating characteristic curve",
        "width": width,
    }

    if type(df_labels_with_splink_scores) == pyspark.sql.DataFrame:
        del roc_chart_def["encoding"]["color"]
        df_labels_with_splink_scores = {"model1": df_labels_with_splink_scores}

    dfs = []
    for key, df in df_labels_with_splink_scores.items():
        data = truth_space_table(
            df, spark, threshold_actual=threshold_actual
        ).toPandas()
        data["roc_label"] = key

        dfs.append(data)

    if not x_domain:

        f1 = data["FP_rate"] < 1.0
        filtered = data[f1]
        d1 = filtered["FP_rate"].max() * 1.5

        x_domain = [0, d1]

    roc_chart_def["encoding"]["x"]["scale"] = {"domain": x_domain}

    records = []
    for df in dfs:
        recs = df.to_dict(orient="records")
        records.extend(recs)

    roc_chart_def["data"]["values"] = records

    if altair_installed:
        return alt.Chart.from_dict(roc_chart_def)
    else:
        return roc_chart_def


def precision_recall_chart(
    df_labels_with_splink_scores,
    spark,
    threshold_actual=0.5,
    domain=None,
    width=400,
    height=400,
):
    """Create a precision recall chart from labelled data

    Args:
        df_labels_with_splink_scores (DataFrame): A dataframe of labels and associated splink scores
            usually the output of the truth.labels_with_splink_scores function
        spark (SparkSession): SparkSession object
        threshold_actual (float, optional): Threshold to use in categorising clerical match
            scores into match or no match. Defaults to 0.5.
        x_domain (list, optional): Domain for x axis. Defaults to None.
        width (int, optional):  Defaults to 400.
        height (int, optional):  Defaults to 400.

    """

    pr_chart_def = {
        "$schema": "https://vega.github.io/schema/vega-lite/v4.8.1.json",
        "config": {"view": {"continuousWidth": 400, "continuousHeight": 300}},
        "data": {"name": "data-fadd0e93e9546856cbc745a99e65285d", "values": None},
        "mark": {"type": "line", "clip": True, "point": True},
        "encoding": {
            "tooltip": [
                {"type": "quantitative", "field": "truth_threshold"},
                {"type": "quantitative", "field": "FP_rate"},
                {"type": "quantitative", "field": "TP_rate"},
                {"type": "quantitative", "field": "TP"},
                {"type": "quantitative", "field": "TN"},
                {"type": "quantitative", "field": "FP"},
                {"type": "quantitative", "field": "FN"},
                {"type": "quantitative", "field": "precision"},
                {"type": "quantitative", "field": "recall"},
            ],
            "x": {
                "type": "quantitative",
                "field": "recall",
                "sort": ["-recall"],
                "title": "Recall",
            },
            "y": {
                "type": "quantitative",
                "field": "precision",
                "sort": ["-precision"],
                "title": "Precision",
            },
        },
        "height": height,
        "title": "Precision-recall curve",
        "width": width,
    }

    if domain:
        pr_chart_def["encoding"]["x"]["scale"]["domain"] = domain

    data = truth_space_table(
        df_labels_with_splink_scores, spark, threshold_actual=threshold_actual
    ).toPandas()

    data = data.to_dict(orient="records")

    pr_chart_def["data"]["values"] = data

    if altair_installed:
        return alt.Chart.from_dict(pr_chart_def)
    else:
        return pr_chart_def



def _sql_expr_move_left_to_right(
    col_name,
    unique_id_col: str = "unique_id",
    source_dataset_col: str = "source_dataset",
):

    sds_l = f"{source_dataset_col}_l"
    uid_l = f"{unique_id_col}_l"
    sds_r = f"{source_dataset_col}_r"
    uid_r = f"{unique_id_col}_r"
    col_name_l = f"{col_name}_l"
    col_name_r = f"{col_name}_r"

    if source_dataset_col:
        uid_expr_l = f"concat({sds_l}, '-__-', {uid_l})"
        uid_expr_r = f"concat({sds_r}, '-__-', {uid_r})"
    else:
        uid_expr_l = uid_l
        uid_expr_r = uid_r

    move_to_left = f"""
    CASE
    WHEN {uid_expr_l} < {uid_expr_r}
    THEN {col_name_l}
    ELSE {col_name_r}
    END as {col_name_l}
    """

    move_to_right = f"""
    CASE
    WHEN {uid_expr_l} < {uid_expr_r}
    THEN {col_name_r}
    ELSE {col_name_l}
    END as {col_name_r}
    """

    exprs = f"""
    {move_to_left},
    {move_to_right}
    """

    return exprs


def lower_id_to_left_hand_side(
    df,
    source_dataset_col: str = "source_dataset",
    unique_id_col: str = "unique_id",
):
    """Take a dataframe in the format of splink record comparisons (with _l and _r suffixes)
    and return a dataframe where the _l columns correspond to the record with the lower id.

    For example:
    | source_dataset_l   |   unique_id_l | source_dataset_r   |   unique_id_r |   a_l |   a_r | other_col   |
    |:-------------------|--------------:|:-------------------|--------------:|------:|------:|:------------|
    | df                 |             0 | df                 |             1 |     0 |     1 | a           |
    | df                 |             2 | df                 |             0 |     2 |     0 | b           |
    | df                 |             0 | df                 |             3 |     0 |     3 | c           |

    Becomes

    | source_dataset_l   |   unique_id_l | source_dataset_r   |   unique_id_r |   a_l |   a_r | other_col   |
    |:-------------------|--------------:|:-------------------|--------------:|------:|------:|:------------|
    | df                 |             0 | df                 |             1 |     0 |     1 | a           |
    | df                 |             0 | df                 |             2 |     0 |     2 | b           |
    | df                 |             0 | df                 |             3 |     0 |     3 | c           |


    Returns:
        df: a dataframe with the columns _l and _r swapped in the case where
            the unique_id_r < unique_id_l
    """
    spark = df.sql_ctx.sparkSession
    cols = list(df.columns)

    l_cols = [c for c in cols if c.endswith("_l")]
    r_cols = [c for c in cols if c.endswith("_r")]
    other_cols = [c for c in cols if c not in (l_cols + r_cols)]

    case_exprs = []
    for col in l_cols:
        this_col = col[:-2]
        expr = _sql_expr_move_left_to_right(this_col, unique_id_col, source_dataset_col)
        case_exprs.append(expr)
    case_exprs.extend(other_cols)
    select_expr = ", ".join(case_exprs)

    df.createOrReplaceTempView("df")
    sql = f"""
    select {select_expr}
    from df
    """

    df = spark.sql(sql)
    return df.select(cols)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['splink']

package_data = \
{'': ['*'],
 'splink': ['files/*',
            'files/chart_defs/*',
            'files/external_js/*',
            'files/templates/*',
            'jars/*']}

install_requires = \
['jsonschema>=3.2,<4.0', 'sqlglot>=1.18.0,<2.0.0', 'typeguard>=2.10.0,<3.0.0']

setup_kwargs = {
    'name': 'splink',
    'version': '2.1.9',
    'description': "Implementation of Fellegi-Sunter's canonical model of record linkage in Apache Spark, including EM algorithm to estimate parameters",
    'long_description': "![image](https://user-images.githubusercontent.com/7570107/85285114-3969ac00-b488-11ea-88ff-5fca1b34af1f.png)\n\n[![Coverage Status](https://coveralls.io/repos/github/moj-analytical-services/splink/badge.svg?branch=master)](https://coveralls.io/github/moj-analytical-services/splink?branch=master)\n![issues-status](https://img.shields.io/github/issues-raw/moj-analytical-services/splink)\n![python-version-dependency](https://img.shields.io/badge/python-%3E%3D3.6-blue)\n\n# splink: Probabilistic record linkage and deduplication at scale\n\n`splink` implements Fellegi-Sunter's canonical model of record linkage in Apache Spark, including the EM algorithm to estimate parameters of the model.\n\nIt:\n\n- Works at much greater scale than current open source implementations (100 million records+).\n\n- Runs quickly - with runtimes of less than an hour.\n\n- Has a highly transparent methodology; match scores can be easily explained both graphically and in words\n\n- Is highly accurate\n\nIt is assumed that users of Splink are familiar with the probabilistic record linkage theory, and the Fellegi-Sunter model in particular. A [series of interactive articles](https://www.robinlinacre.com/probabilistic_linkage/) explores the theory behind Splink.\n\nThe statistical model behind `splink` is the same as that used in the R [fastLink package](https://github.com/kosukeimai/fastLink). Accompanying the fastLink package is an [academic paper](http://imai.fas.harvard.edu/research/files/linkage.pdf) that describes this model. This is the best place to start for users wanting to understand the theory about how `splink` works.\n\n[Data Matching](https://link.springer.com/book/10.1007/978-3-642-31164-2), a book by Peter Christen, is another excellent resource.\n\n## Installation\n\n`splink` is a Python package. It uses the Spark Python API to execute data linking jobs in a Spark cluster. It has been tested in Apache Spark 2.3, 2.4 and 3.1.\n\nInstall splink using:\n\n`pip install splink`\n\nNote that Splink requires `pyspark` and a working Spark installation. These are not specified as explicit dependencies becuase it is assumed users have an existing pyspark setup they wish to use.\n\n## Interactive demo\n\nYou can run demos of `splink` in an interactive Jupyter notebook by clicking the button below:\n\n[![Binder](https://mybinder.org/badge.svg)](https://mybinder.org/v2/gh/moj-analytical-services/splink_demos/master?urlpath=lab/tree/index.ipynb)\n\n## Documentation\n\nThe best documentation is currently a series of demonstrations notebooks in the [splink_demos](https://github.com/moj-analytical-services/splink_demos) repo.\n\n## Other tools in the Splink family\n\n### Splink Graph\n\n[`splink_graph`](https://github.com/moj-analytical-services/splink_graph) is a graph utility library for use in Apache Spark. It computes graph metrics on the outputs of data linking. The repo is [here](<(https://github.com/moj-analytical-services/splink_graph)>)\n\n- Quality assurance of linkage results and identifying false positive links\n- Computing quality metrics associated with groups (clusters) of linked records\n- Automatically identifying possible false positive links in clusters\n\n\n### Splink Comparison Viewer\n\n[`splink_comparison_viewer`](https://github.com/moj-analytical-services/splink_comparison_viewer) produces interactive dashboards that help you rapidly understand and quality assure the outputs of record linkage.    A tutorial video is available [here](https://www.youtube.com/watch?v=DNvCMqjipis).\n\n### Splink Cluster Studio\n\n[`splink_cluster_studio`](http://github.com/moj-analytical-services/splink_cluster_studio) creates an interactive html dashboard from Splink output that allows you to visualise and analyse a sample of clusters from your record linkage. The repo is [here](http://github.com/moj-analytical-services/splink_cluster_studio).\n\n### Splink Synthetic Data\n\nThis [code](https://github.com/moj-analytical-services/splink_synthetic_data) is able to generate realistic test datasets for linkage using the WikiData Query Service.\n\nIt has been used to [performance test the accuracy of various Splink models](https://www.robinlinacre.com/comparing_splink_models/).\n\n### Interactive settings editor with autocomplete\n\nWe also provide an interactive `splink` settings editor and example settings [here](https://moj-analytical-services.github.io/splink_settings_editor/).\n\n### Starting parameter generation tools\n\nA tool to generate custom `m` and `u` probabilities can be found [here](https://www.robinlinacre.com/m_u_generator/).\n\n## Blog\n\nYou can read a short blog post about `splink` [here](https://robinlinacre.com/introducing_splink/).\n\n## Videos\n\nYou can find a short video introducing `splink` and running though an introductory demo [here](https://www.youtube.com/watch?v=_8lV2Lbd6Xs&feature=youtu.be&t=1295).\n\nA 'best practices and performance tuning' tutorial can be found [here](https://www.youtube.com/watch?v=HzcqrRvXhCE).\n\n## How to make changes to Splink\n\n(Steps 5 onwards for repo admins only)\n\n1. Raise new issue or target existing issue\n2. Create new branch (usually off master).  Or fork for external contributors.\n3. Make changes, commit and push to GitHub\n4. Make pull request, referencing the issue\n5. Wait for tests to pass\n6. Review pull request\n7. Bump Splink version in [pyproject.toml](https://github.com/moj-analytical-services/splink/blob/master/pyproject.toml) and update [CHANGELOG.md](https://github.com/moj-analytical-services/splink/blob/master/CHANGELOG.md) as part of pull request\n8. Merge\n9. Release on PyPI and create new tag\n\n## Acknowledgements\n\nWe are very grateful to [ADR UK](https://www.adruk.org/) (Administrative Data Research UK) for providing funding for this work as part of the [Data First](https://www.adruk.org/our-work/browse-all-projects/data-first-harnessing-the-potential-of-linked-administrative-data-for-the-justice-system-169/) project.\n\nWe are also very grateful to colleagues at the UK's Office for National Statistics for their expert advice and peer review of this work.\n",
    'author': 'Robin Linacre',
    'author_email': 'robinlinacre@hotmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/moj-analytical-services/splink',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import logging

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession

from .logging_utils import _format_sql
from .model import Model

logger = logging.getLogger(__name__)


def _sql_gen_new_lambda(table_name="df_intermediate"):

    sql = f"""
    select cast(sum(expected_num_matches)/sum(num_rows) as double) as new_lambda
    from {table_name}
    """

    return sql


def _get_new_lambda(df_intermediate, spark):
    """
    Calculate lambda, as expected proportion of matches
    given current parameter estimates.

    This can then be used in future iterations.
    """
    df_intermediate.createOrReplaceTempView("df_intermediate")
    sql = _sql_gen_new_lambda(table_name="df_intermediate")

    new_lambda = spark.sql(sql).collect()[0][0]
    logger.debug(_format_sql(sql))
    return new_lambda


def _sql_gen_intermediate_pi_aggregate(model, table_name="df_e"):
    """
    This intermediate step is calculated for efficiency purposes.

    In the maximisation step, to compute the new
    pi probability distributions, we need to perform a variety of calculations that can all be derived
    from this intermediate table.

    Without this intermediate table, we'd be repeating these calculations multiple times.
    """
    ccs = model.current_settings_obj.comparison_columns_list

    gamma_cols_expr = ", ".join([cc.gamma_name for cc in ccs])

    sql = f"""
    select {gamma_cols_expr}, sum(match_probability) as expected_num_matches, sum(1- match_probability) as expected_num_non_matches, count(*) as num_rows
    from {table_name}
    group by {gamma_cols_expr}
    """
    return sql


def _sql_gen_pi_df(model, table_name="df_intermediate"):

    sqls = []
    for cc in model.current_settings_obj.comparison_columns_list:
        gamma_column_name = cc.gamma_name
        col_name = cc.name
        sql = f"""
        select
        {gamma_column_name} as gamma_value,
        cast(sum(expected_num_matches)/(select sum(expected_num_matches) from {table_name} where {gamma_column_name} != -1) as double) as new_probability_match,
        cast(sum(expected_num_non_matches)/(select sum(expected_num_non_matches) from {table_name} where {gamma_column_name} != -1) as double) as new_probability_non_match,
        '{col_name}' as column_name
        from {table_name}
        where {gamma_column_name} != -1
        group by {gamma_column_name}
        """
        sqls.append(sql)

    sql = "\nunion all\n".join(sqls)

    return sql


def _get_new_pi_df(df_intermediate, spark, params):
    """
    Calculate and collect a dataframe that contains all the new values of pi
    """
    df_intermediate.createOrReplaceTempView("df_intermediate")
    sql = _sql_gen_pi_df(params)
    levels = spark.sql(sql).collect()
    logger.debug(_format_sql(sql))
    return [l.asDict() for l in levels]


def run_maximisation_step(df_e: DataFrame, model: Model, spark: SparkSession):
    """Compute new parameters and save them in the model object

    Note that the model object will be updated in-place by this function

    Args:
        df_e (DataFrame): the result of the expectation step
        model (Model): splink Model object
        spark (SparkSession): The spark session
    """

    sql = _sql_gen_intermediate_pi_aggregate(model)

    df_e.createOrReplaceTempView("df_e")
    df_intermediate = spark.sql(sql)
    logger.debug(_format_sql(sql))
    df_intermediate.createOrReplaceTempView("df_intermediate")
    df_intermediate.persist()

    new_lambda = _get_new_lambda(df_intermediate, spark)
    pi_df_collected = _get_new_pi_df(df_intermediate, spark, model)

    model._populate_model_from_maximisation_step(new_lambda, pi_df_collected)
    model.iteration += 1
    df_intermediate.unpersist()



from functools import reduce
from copy import copy, deepcopy
import warnings
import pandas as pd

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from typeguard import typechecked

from .charts import load_chart_definition, altair_if_installed_else_json
from .settings import Settings


def _equal_spaced_buckets(num_buckets, extent):
    buckets = [x for x in range(num_buckets + 1)]
    span = extent[1] - extent[0]
    buckets = [extent[0] + span * (x / num_buckets) for x in buckets]
    return buckets


@typechecked
def _calc_probability_density(
    df_e: DataFrame,
    spark: SparkSession,
    buckets=None,
    score_colname="match_probability",
    symmetric=True,
):

    """perform splink score histogram calculations / internal function

    Compute a histogram using the provided buckets.

        Args:
            df_e (DataFrame): A dataframe of record comparisons containing a
                splink score, e.g. as produced by the expectation step
            spark (SparkSession): SparkSession object
            score_colname: is the score in another column? defaults to match_probability.  also try match_weight
            buckets: accepts either a list of split points or an integer number that is used
                to create equally spaced split points.  It defaults to 100 equally
                spaced split points
            symmetric : if True then the histogram is symmetric

        Returns:
            (list) : list of rows of histogram bins for appropriate splink score variable ready to be plotted.
    """

    if score_colname == "match_probability":
        extent = (0.0, 1.0)
    else:
        weight_max = df_e.agg({score_colname: "max"}).collect()[0][0]
        weight_min = df_e.agg({score_colname: "min"}).collect()[0][0]
        extent = (weight_min, weight_max)
        if symmetric:
            extent_max = max(abs(weight_max), abs(weight_min))
            extent = (-extent_max, extent_max)

    # if buckets a list then use it. if None... then create default. if integer then create equal bins
    if isinstance(buckets, int) and buckets != 0:
        buckets = _equal_spaced_buckets(buckets, extent)
    elif buckets is None:
        buckets = _equal_spaced_buckets(100, extent)

    buckets.sort()

    # ensure bucket splits are in ascending order
    if score_colname == "match_probability":
        if buckets[0] != 0:
            buckets = [0.0] + buckets

        if buckets[-1] != 1.0:
            buckets = buckets + [1.0]

    hist = df_e.select(score_colname).rdd.flatMap(lambda x: x).histogram(buckets)

    # get bucket from and to points
    bin_low = hist[0]
    bin_high = copy(hist[0])
    bin_low.pop()
    bin_high.pop(0)
    counts = hist[1]

    rows = []
    for item in zip(bin_low, bin_high, counts):
        new_row = {
            "splink_score_bin_low": item[0],
            "splink_score_bin_high": item[1],
            "count_rows": item[2],
        }
        rows.append(new_row)

    for r in rows:
        r["binwidth"] = r["splink_score_bin_high"] - r["splink_score_bin_low"]
        r["freqdensity"] = r["count_rows"] / r["binwidth"]

    sumfreqdens = reduce(lambda a, b: a + b["freqdensity"], rows, 0)

    for r in rows:
        r["normalised"] = r["freqdensity"] / sumfreqdens

    return rows


def _create_probability_density_plot(data):
    """plot score histogram

    Args:
        data (list): A list of rows of histogram bins
            as produced by the _calc_probability_density function
    Returns:
        if altair is installed a plot. if altair is not installed
            then it returns the vega lite chart spec as a dictionary
    """

    hist_def_dict = load_chart_definition("score_histogram.json")
    hist_def_dict["data"]["values"] = data

    return altair_if_installed_else_json(hist_def_dict)


def splink_score_histogram(
    df_e: DataFrame,
    spark: SparkSession,
    buckets=None,
    score_colname="match_weight",
    symmetric=True,
):

    """splink score histogram diagnostic plot public API function

    Compute a histogram using the provided buckets and plot the result.

    Args:
        df_e (DataFrame): A dataframe of record comparisons containing a splink score,
            e.g. as produced by the `get_scored_comparisons` function
        spark (SparkSession): SparkSession object
        score_colname : is the score in another column? defaults to match_weight.  also try match_probability
        buckets : accepts either a list of split points or an integer number that is used to
            create equally spaced split points. It defaults to 100 equally spaced split points from 0.0 to 1.0
        symmetric : if True then the histogram is symmetric
     Returns:
        if altair library is installed this function returns a histogram plot. if altair is not installed
        then it returns the vega lite chart spec as a dictionary
    """

    rows = _calc_probability_density(
        df_e,
        spark=spark,
        buckets=buckets,
        score_colname=score_colname,
        symmetric=symmetric,
    )

    return _create_probability_density_plot(rows)


def comparison_vector_distribution(df_gammas):
    """Generate a pandas dataframe containing the
    distribution of counts of distinct comparison vectors
    in df_gammas or df_e.

    If df_e is provided, the average match_probability
    and match_weight will be output for each comparison vector.

    Args:
        df_gammas (DataFrame): A dataframe of record comparisons
    """

    spark = df_gammas.sql_ctx.sparkSession

    df_gamma_cols = df_gammas.columns

    g_cols = [c for c in df_gamma_cols if c.startswith("gamma_")]
    sel_cols = g_cols.copy()

    if "match_probability" in df_gamma_cols:
        sel_cols.append("match_probability")
    if "match_weight" in df_gamma_cols:
        sel_cols.append("match_weight")

    df_gammas = df_gammas.select(sel_cols)

    cols_expr = ", ".join([f'"{c}"' for c in g_cols])
    cols_expr = ", ".join(g_cols)

    df_gammas.createOrReplaceTempView("df_gammas")

    case_tem = "(case when {g} = -1 then 0 when {g} = 0 then -1 else {g} end)"
    sum_gams = " + ".join([case_tem.format(g=c) for c in g_cols])

    score_col_expr = []
    if "match_probability" in df_gamma_cols:
        score_col_expr.append(f"avg(match_probability) as match_probability")
    if "match_weight" in df_gamma_cols:
        score_col_expr.append(f"avg(match_weight) as match_weight")
    if score_col_expr:
        score_col_expr = ", " + ", ".join(score_col_expr)
    else:
        score_col_expr = ""

    sql = f"""
    select {cols_expr}, concat_ws(',', {cols_expr}) as gam_concat, {sum_gams} as sum_gam, count(*) as count {score_col_expr}
    from df_gammas
    group by {cols_expr}
    order by {cols_expr}
    """

    gamma_counts = spark.sql(sql)
    gamma_counts = gamma_counts.toPandas()
    gamma_counts["proportion_of_comparisons"] = (
        gamma_counts["count"] / gamma_counts["count"].sum()
    )

    gamma_counts = gamma_counts.sort_values("sum_gam")
    if "match_probability" in gamma_counts.columns:
        gamma_counts = gamma_counts.sort_values("match_probability")

    if "match_weight" in gamma_counts.columns:
        gamma_counts = gamma_counts.sort_values("match_weight")

    gamma_counts["cumulative_comparisons"] = gamma_counts[
        "proportion_of_comparisons"
    ].cumsum()

    return gamma_counts


COLOUR_ENCODING_MATCH_PROB = {
    "type": "quantitative",
    "field": "match_probability",
    "scale": {"range": ["red", "orange", "green"], "domain": [0, 0.5, 1]},
}

COLOUR_ENCODING_MATCH_WEIGHT = {
    "type": "quantitative",
    "field": "match_weight",
    "scale": {
        "range": ["red", "red", "orange", "green", "green"],
        "domain": [-100, -10, 0, 10, 100],
    },
}


def comparison_vector_distribution_chart(cvd_df, symlog=True, symlog_constant=40):
    """Generate an Altair chart from the comparison vector distribution
    produced by splink.diagnostics.comparison_vector_distribution()

    Args:
        cvd_df (pandas.DataFrame): Output of splink.diagnostics.comparison_vector_distribution()
        symlog (bool, optional): If true, y axis is symlog. Defaults to True.
        symlog_constant (int, optional): Symlog constant if symlog set to True. Defaults to 40.
    """

    hist_def_dict = load_chart_definition("gamma_histogram.json")
    hist_def_dict["data"]["values"] = cvd_df.to_dict(orient="records")

    if not symlog:
        del hist_def_dict["encoding"]["y"]["scale"]
    else:
        hist_def_dict["encoding"]["y"]["scale"]["constant"] = symlog_constant

    tooltips = [
        {"field": "gam_concat", "type": "nominal"},
        {"field": "count", "type": "quantitative"},
        {"field": "proportion_of_comparisons", "type": "quantitative", "format": ".2%"},
        {"field": "cumulative_comparisons", "type": "quantitative", "format": ".2%"},
    ]

    cvd_columns = list(cvd_df.columns)
    score_tts = []
    if "match_probability" in cvd_columns:
        score_tts.append(
            {"field": "match_probability", "type": "quantitative", "format": ".2%"}
        )
    if "match_weight" in cvd_columns:
        score_tts.append(
            {"field": "match_weight", "type": "quantitative", "format": ".2f"}
        )
    if len(score_tts) == 0:
        score_tts.append({"field": "sum_gam", "type": "quantitative"})

    tooltips.extend(score_tts)
    g_cols = [c for c in cvd_df.columns if c.startswith("gamma_")]
    g_tts = [{"field": c, "type": "nominal"} for c in g_cols]
    tooltips.extend(g_tts)

    hist_def_dict["encoding"]["tooltip"] = tooltips

    hist_def_dict["encoding"]["x"]["sort"]["field"] = "sum_gam"

    if "match_weight" in cvd_columns:
        hist_def_dict["encoding"]["color"] = COLOUR_ENCODING_MATCH_WEIGHT
        hist_def_dict["encoding"]["x"]["sort"]["field"] = "match_weight"

    if "match_probability" in cvd_columns:
        hist_def_dict["encoding"]["color"] = COLOUR_ENCODING_MATCH_PROB
        hist_def_dict["encoding"]["x"]["sort"]["field"] = "match_probability"

    return altair_if_installed_else_json(hist_def_dict)


def _melted_comparison_vector_distribution(cvd):

    cvd = cvd.reset_index()
    cvd = cvd.rename(columns={"index": "comparison_vector_uid"})

    # Pivot cvd2 from wide format to long format on gamma cols
    all_cols = cvd.columns.tolist()
    index_cols = [c for c in all_cols if not c.startswith("gamma_")]

    cvd_melted = cvd.melt(id_vars=index_cols).sort_values("comparison_vector_uid")
    cvd_melted = cvd_melted.rename(
        columns={"value": "gam_val", "variable": "gam_colname"}
    )
    return cvd_melted


def _m_u_table_with_null_adjustment(null_props, settings, spark):
    settings_obj = Settings(settings)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        settings_obj.complete_settings_dict(spark)
    out_data = []
    for cc in settings_obj.comparison_columns_list:
        gam_name = cc.gamma_name
        null_prop = null_props[gam_name]
        populated_prop = 1 - null_prop
        row = {
            "gam_val": -1,
            "gam_colname": gam_name,
            "m_probability": null_prop,
            "u_probability": null_prop,
        }
        out_data.append(row)
        for row in cc.as_rows():
            m = row["m_probability"]
            u = row["u_probability"]
            gam_val = row["gamma_index"]

            row = {
                "gam_val": gam_val,
                "gam_colname": gam_name,
                "m_probability": m * populated_prop,
                "u_probability": u * populated_prop,
            }
            out_data.append(row)
    prob_lookup = pd.DataFrame(out_data)
    return prob_lookup


def get_theoretical_comparison_vector_distribution(df_gammas, actual_cvd, settings):
    """Generate a pandas dataframe containing the distribution of
    distinct comparison vectors that would be expected
    from the m and u values in the settings object if the assumptions
    of the fellegi sunter model held (i.e. conditional independence
    of comparison vector values given match status).
    Only produces output for the patterns of comparison vectors
    in df_gammas

    If df_e is provided, the average match_probability
    and match_weight will be output for each comparison vector.

    Args:
        df_gammas (DataFrame): A dataframe of record comparisons
    """

    spark = df_gammas.sql_ctx.sparkSession

    settings_obj = Settings(settings)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        settings_obj.complete_settings_dict(spark)

    lam = settings_obj["proportion_of_matches"]

    total_cvs = actual_cvd["count"].sum()
    cvd_melted = _melted_comparison_vector_distribution(actual_cvd)

    # df gammas is needed to estimate the proportion of record comparisons that are null
    null_props = estimate_proportion_of_null_comparisons(df_gammas, settings)
    m_u_lookup = _m_u_table_with_null_adjustment(null_props, settings, spark)

    df_cvd_with_m_u = cvd_melted.merge(
        m_u_lookup,
        left_on=["gam_colname", "gam_val"],
        right_on=["gam_colname", "gam_val"],
    )

    gamma_cols = df_gammas.columns
    actual_cvd_cols = list(actual_cvd.columns)

    index_cols = ["comparison_vector_uid", "gam_concat", "sum_gam"]

    if "match_weight" in gamma_cols and "match_weight" in actual_cvd_cols:
        index_cols.append("match_weight")
    if "match_probability" in gamma_cols and "match_probability" in actual_cvd_cols:
        index_cols.append("match_probability")

    pt1 = df_cvd_with_m_u.pivot_table(
        index=index_cols,
        values=["u_probability", "m_probability"],
        aggfunc=pd.Series.product,
    )
    pt1["proportion_of_comparisons"] = (1 - lam) * pt1["u_probability"] + lam * pt1[
        "m_probability"
    ]

    pt2 = df_cvd_with_m_u.pivot_table(
        index=index_cols, columns="gam_colname", values="gam_val", aggfunc="mean"
    )

    final = pt1.join(pt2).reset_index()
    final["proportion_of_comparisons"] = (
        final["proportion_of_comparisons"] / final["proportion_of_comparisons"].sum()
    )
    final["count"] = final["proportion_of_comparisons"] * total_cvs
    final = final.drop(
        ["comparison_vector_uid", "m_probability", "u_probability"], axis=1
    )

    final = final.sort_values("sum_gam")
    if "match_probability" in final.columns:
        final = final.sort_values("match_probability")

    if "match_weight" in final.columns:
        final = final.sort_values("match_weight")

    final["cumulative_comparisons"] = final["proportion_of_comparisons"].cumsum()

    return final


def estimate_proportion_of_null_comparisons(
    df_gammas: DataFrame,
    settings: dict,
):

    spark = df_gammas.sql_ctx.sparkSession

    settings_obj = Settings(settings)

    gamma_cols = [cc.gamma_name for cc in settings_obj.comparison_columns_list]

    sql_template = """

    select count(*)/(select count(*) from df_gammas) as proportion, {gamma_name} as gam_val, '{gamma_name}' as gam_colname
    from df_gammas
    group by {gamma_name}
    """

    sqls = [sql_template.format(gamma_name=c) for c in gamma_cols]

    unions = " UNION ALL ".join(sqls)

    df_gammas.createOrReplaceTempView("df_gammas")
    sql = f"""

    with unions as ({unions})

    select proportion as null_proportion, gam_colname from unions where gam_val = -1

    """
    null_tab = spark.sql(sql).toPandas()
    null_proportions = null_tab.set_index("gam_colname").to_dict(orient="index")

    result = {k: v["null_proportion"] for k, v in null_proportions.items()}

    for c in gamma_cols:
        if c not in result:
            result[c] = 0.0

    return result


def compare_actual_and_theoretical_cvd(actual_cvd, theoretical_cvd):
    """Compare actual to theoretical distribution

    Args:
        actual_cvd: Output of comparison_vector_distribution()
        theoretical_cvd: Output of  get_theoretical_comparison_vector_distribution()

    Returns:
        Pandas dataframe comparing actual to theoretical distribution
    """
    theoretical_cvd = theoretical_cvd[["gam_concat", "count"]].copy()
    theoretical_cvd = theoretical_cvd.rename(columns={"count": "count_theoretical"})

    actual_cvd = actual_cvd.rename(columns={"count": "count_actual"})

    merged = actual_cvd.merge(
        theoretical_cvd, left_on="gam_concat", right_on="gam_concat"
    )

    merged["count_diff"] = merged["count_actual"] - merged["count_theoretical"]
    return merged


def comparison_vector_comparison_chart(
    cvd_comparison_df, symlog=True, symlog_constant=40
):
    """Generate an Altair chart that compares actual
    to theoretical distribution

    Args:
        cvd_df (pandas.DataFrame): Output of splink.diagnostics.compare_actual_and_theoretical_cvd()
        symlog (bool, optional): If true, y axis is symlog. Defaults to True.
        symlog_constant (int, optional): Symlog constant if symlog set to True. Defaults to 40.
    """

    hist_def_dict = load_chart_definition("gamma_histogram.json")
    hist_def_dict["data"]["values"] = cvd_comparison_df.to_dict(orient="records")
    hist_def_dict["encoding"]["y"]["field"] = "count_diff"

    if not symlog:
        del hist_def_dict["encoding"]["y"]["scale"]
    else:
        hist_def_dict["encoding"]["y"]["scale"]["constant"] = symlog_constant

    tooltips = [
        {"field": "gam_concat", "type": "nominal"},
        {"field": "count_actual", "type": "quantitative"},
        {"field": "count_theoretical", "type": "quantitative"},
        {"field": "count_diff", "type": "quantitative"},
    ]

    cvd_columns = list(cvd_comparison_df.columns)

    score_tts = []
    if "match_probability" in cvd_columns:
        score_tts.append(
            {"field": "match_probability", "type": "quantitative", "format": ".2%"}
        )
    if "match_weight" in cvd_columns:
        score_tts.append(
            {"field": "match_weight", "type": "quantitative", "format": ".2f"}
        )
    if len(score_tts) == 0:
        score_tts.append({"field": "sum_gam", "type": "quantitative"})

    tooltips.extend(score_tts)
    g_cols = [c for c in cvd_comparison_df.columns if c.startswith("gamma_")]
    g_tts = [{"field": c, "type": "nominal"} for c in g_cols]
    tooltips.extend(g_tts)

    hist_def_dict["encoding"]["tooltip"] = tooltips

    hist_def_dict["encoding"]["x"]["sort"]["field"] = "sum_gam"

    if "match_weight" in cvd_columns:
        hist_def_dict["encoding"]["color"] = COLOUR_ENCODING_MATCH_WEIGHT
        hist_def_dict["encoding"]["x"]["sort"]["field"] = "match_weight"
    if "match_probability" in cvd_columns:
        hist_def_dict["encoding"]["color"] = COLOUR_ENCODING_MATCH_PROB
        hist_def_dict["encoding"]["x"]["sort"]["field"] = "match_probability"

    return altair_if_installed_else_json(hist_def_dict)



# For more information on where formulas came from, see
# https://github.com/moj-analytical-services/splink/pull/107

import logging

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession

from .logging_utils import _format_sql
from .model import Model

logger = logging.getLogger(__name__)


def sql_gen_term_frequencies(column_name, table_name="df"):

    sql = f"""
    select
    {column_name}, count(*) / (select
        count({column_name}) as total from {table_name}) as tf_{column_name}
    from {table_name}
    where {column_name} is not null
    group by {column_name}
    """

    return sql


def _sql_gen_add_term_frequencies(model: Model, table_name: str = "df"):
    """Build SQL statement that adds gamma columns to the comparison dataframe

    Args:
        settings (dict): `splink` settings dict
        table_name (str, optional): Name of the source df. Defaults to "df".

    Returns:
        str: A SQL string
    """

    cc_dict = model.current_settings_obj.comparison_column_dict

    cols = [name for name, cc in cc_dict.items() if cc.term_frequency_adjustments]
    # cols = [cc.name for cc in settings["comparison_columns"] if cc["term_frequency_adjustments"]]
    tf_tables = ", ".join(
        [f"tf_{col} as ({sql_gen_term_frequencies(col, table_name)})" for col in cols]
    )

    tf_cols = ", ".join(f"tf_{col}.tf_{col}" for col in cols)

    joins = "".join(
        [
            f"""
    left join tf_{col}
    on {table_name}.{col} = tf_{col}.{col}
    """
            for col in cols
        ]
    )

    sql = f"""
    with {tf_tables}
    select {table_name}.*, {tf_cols}
    from {table_name}
    {joins}
    """

    return sql


def add_term_frequencies(df: DataFrame, model: Model, spark: SparkSession):
    """Compute the term frequencies of the required columns and add to the dataframe.
    Args:
        df (spark dataframe): A Spark dataframe containing source records for linking
        settings_dict (dict): The `splink` settings dictionary
        spark (Spark session): The Spark session object

    Returns:
        Spark dataframe: A dataframe containing new columns representing the term frequencies
        of the corresponding values
    """

    if model.current_settings_obj.any_cols_have_tf_adjustments:
        sql = _sql_gen_add_term_frequencies(model, "df")

        logger.debug(_format_sql(sql))
        df.createOrReplaceTempView("df")
        df_with_tf = spark.sql(sql)

        return df_with_tf
    else:
        return df



from typing import Callable
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession

from .expectation_step import run_expectation_step
from .maximisation_step import run_maximisation_step
from .model import Model
from typeguard import typechecked

import logging

logger = logging.getLogger(__name__)


@typechecked
def iterate(
    df_gammas: DataFrame,
    model: Model,
    spark: SparkSession,

    save_state_fn: Callable = None,
):
    """Repeatedly run expectation and maximisation step until convergence or max itations is reached.

    Args:
        df_gammas (DataFrame): Spark dataframe including gamma columns (i.e. after blocking and add_gammas has been applied)
        model (Model): The `splink` model object
        spark (SparkSession): The SparkSession object
        log_iteration (bool, optional): Whether to write a message to the log after each iteration. Defaults to False.
        num_iterations (int, optional): The number of iterations to run. Defaults to 10.

        save_state_fn (function, optional):  A function provided by the user that takes one arguments, params, and is executed each iteration.  This is a hook that allows the user to save the state between iterations, which is mostly useful for very large jobs which may need to be restarted from where they left off if they fail.

    Returns:
        DataFrame: A spark dataframe including a match probability column
    """
    settings = model.current_settings_obj.settings_dict
    num_iterations = settings["max_iterations"]
    for i in range(num_iterations):

        df_e = run_expectation_step(df_gammas, model, spark)

        run_maximisation_step(df_e, model, spark)

        logger.info(f"Iteration {i} complete")

        if save_state_fn:
            save_state_fn(model)
        if model.is_converged():
            logger.info("EM algorithm has converged")
            break

    # The final version of df_e should align to the current parameters - i.e. those computed in the last max step
    df_e = run_expectation_step(df_gammas, model, spark)

    # The expectation step adds the current params to history, so this is needed to output a final
    # version of charts/params.
    if save_state_fn:
        save_state_fn(model)

    return df_e



from copy import deepcopy
import re

from splink.blocking import _sql_gen_where_condition, block_using_rules


def _get_queryplan_text(df, blocking_rule, splink_settings):
    spark = df.sql_ctx.sparkSession

    # Temporarily override broadcast join threshold to ensure
    # that the query plan is a SortMergeJoin
    current_broadcast_setting = spark.conf.get("spark.sql.autoBroadcastJoinThreshold")
    spark.conf.set("spark.sql.autoBroadcastJoinThreshold", "-1")

    link_type = splink_settings["link_type"]
    if link_type == "dedupe_only":
        source_dataset_col = None
    else:
        source_dataset_col = splink_settings["source_dataset_column_name"]

    unique_id_col = splink_settings["unique_id_column_name"]

    join_filter = _sql_gen_where_condition(link_type, source_dataset_col, unique_id_col)

    df.createOrReplaceTempView("df")
    sql = f"""
    select * from
    df as l
    inner join df as r
    on {blocking_rule}  {join_filter}
    """

    df = spark.sql(sql)
    qe = df._jdf.queryExecution()
    sp = qe.sparkPlan()
    treestring = sp.treeString()
    spark.conf.set("spark.sql.autoBroadcastJoinThreshold", current_broadcast_setting)
    return treestring


def _get_join_line(queryplan_text):
    lines = queryplan_text.splitlines()
    return lines[0]


def _parse_join_line_sortmergejoin(join_line):
    parts = _split_by_commas_ignoring_within_brackets(join_line)
    hash_columns = _get_hash_columns(join_line)

    if len(parts) == 4:
        post_join_filters = _remove_col_ids(parts[3])
    else:
        post_join_filters = None

    return {
        "join_strategy": "SortMergeJoin",
        "join_type": "Inner",
        "join_hashpartition_columns_left": hash_columns["left"],
        "join_hashpartition_columns_right": hash_columns["right"],
        "post_join_filters": post_join_filters,
    }


def _parse_join_line_cartesian(join_line):
    fil = _extract_text_from_within_brackets_balanced(join_line, ["(", ")"])
    fil = _remove_col_ids(fil)
    return {
        "join_strategy": "Cartesian",
        "join_type": "Cartesian",
        "join_hashpartition_columns_left": [],
        "join_hashpartition_columns_right": [],
        "post_join_filters": fil,
    }


def _parse_join_line(join_line):
    if "SortMergeJoin" in join_line:
        return _parse_join_line_sortmergejoin(join_line)
    if "Cartesian" in join_line:
        return _parse_join_line_cartesian(join_line)


def _split_by_commas_ignoring_within_brackets(input_str):
    counter = 0
    captured_strings = []
    captured_string = ""
    for i in input_str:
        captured_string += i

        if i in ("(", "[", "{"):
            counter += 1

        if i in (")", "]", "}"):
            counter -= 1
        if counter == 0 and i == ",":
            captured_string = captured_string[:-1]
            captured_strings.append(captured_string)
            captured_string = ""
    captured_strings.append(captured_string)
    captured_strings = [s.strip() for s in captured_strings]
    return captured_strings


def _extract_text_from_within_brackets_balanced(input_str, bracket_type=["[", "]"]):
    bracket_counter = 0

    start_bracket = bracket_type[0]
    end_bracket = bracket_type[1]

    captured_string = ""

    if start_bracket not in input_str:
        return None

    for i in input_str:

        if i == start_bracket:
            bracket_counter += 1

        if i == end_bracket:
            bracket_counter -= 1
            if bracket_counter == 0:
                break
        if bracket_counter > 0:
            captured_string += i

    return captured_string[1:]


def _remove_col_ids(input_str):
    return re.sub(r"#\d{1,6}[L]?", "", input_str)


def get_total_comparisons_from_join_columns_that_will_be_hash_partitioned(
    df, join_cols: list
):
    """Compute the total number of records that will be generated
    by an inner joi on the join_cols
    For instance, if join_cols = ["first_name", "dmetaphone(surname)"]
    will compute the total number of comparisons generated by the blocking rule:
    ["l.first_name = r.first_name and dmetaphone(l.surname) = dmetaphone(r.surname)"]
    Args:
        df (DataFrame): Input dataframe
        join_cols (list): List of blocking columns e.g. ["first_name", "dmetaphone(surname)"]
    Returns:
        integer: Number of comparisons generated by the blocking rule
    """

    sel_expr = ", ".join(join_cols)
    concat_expr = f"concat({sel_expr})"
    spark = df.sql_ctx.sparkSession

    df.createOrReplaceTempView("df")

    sql = f"""
    with
    block_groups as (
        SELECT {concat_expr}, {sel_expr},
        count(*) * count(*) as num_comparisons
    FROM df
    where {concat_expr} is not null
    GROUP BY {concat_expr}, {sel_expr}
    )
    select sum(num_comparisons) as total_comparisons
    from block_groups
    """

    return spark.sql(sql).collect()[0]["total_comparisons"]


def generate_and_count_num_comparisons_from_blocking_rule(
    df, blocking_rule, splink_settings
):
    spark = df.sql_ctx.sparkSession
    splink_settings = deepcopy(splink_settings)
    splink_settings["blocking_rules"] = [blocking_rule]

    df = block_using_rules(splink_settings, df, spark)

    return df.count()


def _get_hash_columns(smj_line):
    comma_split = _split_by_commas_ignoring_within_brackets(smj_line)
    join_left = _extract_text_from_within_brackets_balanced(comma_split[0])
    join_left = _remove_col_ids(join_left)
    join_left = _split_by_commas_ignoring_within_brackets(join_left)

    join_right = _extract_text_from_within_brackets_balanced(comma_split[1])
    join_right = _remove_col_ids(join_right)
    join_right = _split_by_commas_ignoring_within_brackets(join_right)
    return {"left": join_left, "right": join_right}


def _get_largest_group(df, join_cols):

    sel_expr = ", ".join(join_cols)
    concat_expr = f"concat({sel_expr})"
    concat_ws_expr = f"concat_ws('|', {sel_expr})"
    spark = df.sql_ctx.sparkSession

    df.createOrReplaceTempView("df")

    sql = f"""
    SELECT {concat_expr} as concat_expr,  {concat_ws_expr} as concat_ws_expr, {sel_expr},
        count(*) * count(*) as num_comparisons
    FROM df
    where {concat_expr} is not null
    GROUP BY {concat_expr}, {sel_expr}
    ORDER BY count(*) desc
    limit 1
    """

    collected = spark.sql(sql).collect()
    largest_group_comparisons = collected[0]["num_comparisons"]
    largest_group_concat = collected[0]["concat_ws_expr"]
    return {
        "largest_group_expr": largest_group_concat,
        "comparisons_generated_in_largest_group_before_filter_applied": largest_group_comparisons,
    }


def analyse_blocking_rule(
    df,
    blocking_rule,
    splink_settings,
    compute_exact_comparisons=False,
    compute_exact_limit=1e9,
    compute_largest_group=True,
):
    spark = df.sql_ctx.sparkSession
    df.createOrReplaceTempView("df")
    splink_settings = deepcopy(splink_settings)
    splink_settings["additional_columns_to_retain"] = []
    queryplan_text = _get_queryplan_text(df, blocking_rule, splink_settings)
    results_dict = _parse_join_line(_get_join_line(queryplan_text))

    # Message detailing any metrics that could not be computed
    metric_message = ""

    if results_dict["join_strategy"] == "SortMergeJoin":

        jcl = results_dict["join_hashpartition_columns_left"]
        jcr = results_dict["join_hashpartition_columns_right"]
        balanced_join = jcl == jcr
        if balanced_join:
            total_comparisons_generated = (
                get_total_comparisons_from_join_columns_that_will_be_hash_partitioned(
                    df, jcl
                )
            )
            results_dict[
                "total_comparisons_generated_before_filter_applied"
            ] = total_comparisons_generated
            if compute_largest_group:
                group_stats = _get_largest_group(df, jcl)
                results_dict = {**results_dict, **group_stats}
        else:
            metric_message += "Join columns include inversions, so total comparisons and largest group could not be computed efficiently.\n"
            results_dict["total_comparisons_generated_before_filter_applied"] = None
            total_comparisons_generated = 1
            if compute_largest_group:
                results_dict["largest_group_expr"] = None
                results_dict[
                    "comparisons_generated_in_largest_group_before_filter_applied"
                ] = None

    if results_dict["join_strategy"] == "Cartesian":
        raw_count = df.count()
        total_comparisons_generated = raw_count * raw_count
        results_dict[
            "total_comparisons_generated_before_filter_applied"
        ] = total_comparisons_generated
        if compute_largest_group:
            results_dict["largest_group_expr"] = None
            results_dict[
                "comparisons_generated_in_largest_group_before_filter_applied"
            ] = None
            metric_message += "Blocking rule results in Cartesian join so largest groups not applicable.\n"

    if compute_exact_comparisons and total_comparisons_generated < compute_exact_limit:
        splink_settings["blocking_rules"] = [blocking_rule]
        blocked = block_using_rules(splink_settings, df, spark)
        total_with_filters = blocked.count()
        results_dict[
            "total_comparisons_generated_after_filters_applied"
        ] = total_with_filters
    else:
        results_dict["total_comparisons_generated_after_filters_applied"] = None
        if not compute_exact_comparisons:
            metric_message += (
                "Exact number of filtered comparisons not computed."
                " Set compute_exact_comparisons=True to compute.\n"
            )
        if (
            compute_exact_comparisons
            and total_comparisons_generated > compute_exact_limit
        ):
            metric_message += (
                f"Exact number of filtered comparisons not computed because total number"
                f" of comparisons ({total_comparisons_generated:,.0f}) exceeds limit"
                f" ({compute_exact_limit:,.0f}).  Set compute_exact_limit to a larger value to compute."
            )
    results_dict["message"] = metric_message

    results_order = [
        "total_comparisons_generated_before_filter_applied",
        "total_comparisons_generated_after_filters_applied",
        "largest_group_expr",
        "comparisons_generated_in_largest_group_before_filter_applied",
        "message",
        "join_strategy",
        "join_type",
        "join_hashpartition_columns_left",
        "join_hashpartition_columns_right",
        "post_join_filters",
    ]

    results = {}
    for key in results_order:
        results[key] = results_dict[key]

    return results
